import collections
import json
import pathlib
from typing import Union

from .misspelling_checker import MisspellingChecker


class MisspellingJSONDetector(MisspellingChecker):
    """
    Detects misspelled words in files based on a JSON file.
    """

    def __init__(
        self,
        misspelling_json_file: Union[pathlib.Path, str],
    ) -> None:
        """
        Initialises a MisspellingJSONDetector instance.

        Args:
          misspelling_json_file: JSON filename of misspelled words and their corrections.

        Raises:
          IOError: Raised if misspelling_json_file can't be found.
          ValueError: Raised if misspelling_json_file isn't correctly formatted.

        """
        self._misspelling_dict = collections.defaultdict(list)
        with open(misspelling_json_file, "r", encoding="utf-8") as custom_json_file:
            custom_dict_with_misspelled_words = json.load(custom_json_file)
            for (
                bad_word,
                correction,
            ) in custom_dict_with_misspelled_words.items():
                self._misspelling_dict[bad_word].append(correction[0])



import sys
from typing import List

from .words import get_a_line


class SuggestionGenerator:
    """Class to query which correction should be used."""

    def __init__(self):
        self.last_suggestions = {}

    def get_suggestion(self, filename: str, lineno: int, word: str, suggestions: List[str]) -> str:
        """Show line from file, a misspelled word and request replacement."""
        if word not in self.last_suggestions:
            self.last_suggestions[word] = suggestions[0]

        line = get_a_line(filename, lineno)
        list_suggestions = ",".join(suggestions)
        last_suggestion = self.last_suggestions[word]
        sys.stdout.write(f'> {line}\nReplace "{word}" with one of {list_suggestions}\nChoose [{last_suggestion}]:')

        suggestion = sys.stdin.readline().strip()
        if not suggestion:
            suggestion = self.last_suggestions[word]
        else:
            self.last_suggestions[word] = suggestion

        return suggestion



import collections
import pathlib
from typing import Union

from .misspelling_checker import MisspellingChecker


class MisspellingFileDetector(MisspellingChecker):
    """
    Detects misspelled words in files.
    """

    def __init__(
        self,
        misspelling_file: Union[pathlib.Path, str],
    ) -> None:
        """
        Initialises a MisspellingFileDetector instance.

        Args:
          misspelling_file: Filename with a list of misspelled words and their corrections.

        Raises:
          IOError: Raised if misspelling_file can't be found.
          ValueError: Raised if misspelling_file isn't correctly formatted.
        """
        self._misspelling_dict = collections.defaultdict(list)
        with open(misspelling_file, "r", encoding="utf-8") as f:
            for line in f:
                bad_word, correction = line.strip().split(" ", 1)
                self._misspelling_dict[bad_word].append(correction)



from .argument_parser import MisspellingArgumentParser
from .files import expand_directories, parse_file_list
from .suggestions import SuggestionGenerator
from .version import get_version
from .words import esc_file, esc_sed, get_a_line, normalize, same_case, split_words

__all__ = [
    "MisspellingArgumentParser",
    "esc_file",
    "esc_sed",
    "get_a_line",
    "get_version",
    "normalize",
    "same_case",
    "SuggestionGenerator",
    "split_words",
    "expand_directories",
    "parse_file_list",
]



import abc
import pathlib
from codecs import StreamWriter
from typing import DefaultDict, Dict, List, TextIO, Tuple, Union

from tap.tap import TapType

from .utils import SuggestionGenerator


class IMisspellingChecker(metaclass=abc.ABCMeta):
    _misspelling_dict: Union[DefaultDict, Dict]
    suggestion_generator = SuggestionGenerator()

    @abc.abstractmethod
    def check(self, filename: str) -> Tuple[List[Exception], List[List[Union[str, int, str]]]]:
        raise NotImplementedError

    @abc.abstractmethod
    def get_suggestions(self, word: str) -> List[str]:
        raise NotImplementedError

    @abc.abstractmethod
    def dump_corrections(self) -> List[List[str]]:
        raise NotImplementedError

    @abc.abstractmethod
    def print_result(
        self,
        filenames: List[pathlib.Path],
        output: StreamWriter,
    ) -> bool:
        raise NotImplementedError

    @abc.abstractmethod
    def export_result_to_file(self, filenames: List[pathlib.Path], output: TextIO) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def output_sed_commands(
        self,
        parser: TapType,
        args: TapType,
        filenames: List[pathlib.Path],
    ) -> None:
        raise NotImplementedError



import json
import os
import pathlib
from typing import List

from .misspelling_checker import MisspellingChecker


class MisspellingDetector(MisspellingChecker):
    """
    Detects misspelled words in files.
    """

    def __init__(self) -> None:
        """
        Initialises a MisspellingDetector instance using the default json files.
        """
        self._misspelling_dict = {}
        file_paths = self._get_default_json_files()
        for dictionary in file_paths:
            with open(os.path.join(os.path.dirname(__file__), dictionary), encoding="utf-8") as input_file:
                self._misspelling_dict.update(json.load(input_file))

    @staticmethod
    def _get_default_json_files() -> List[pathlib.Path]:
        assets_dir = pathlib.Path(__file__).parents[0] / "assets"
        file_paths = [
            assets_dir.joinpath(file)
            for file in os.listdir(assets_dir.as_posix())
            if os.path.isfile(assets_dir.joinpath(file))
        ]
        return file_paths



import re
import string
from typing import List

_NORM_REGEX = re.compile(r"(?<=[a-z])(?=[A-Z])")
_WORD_REGEX = re.compile(r"[\s_0-9\W]+", flags=re.UNICODE)


def normalize(word: str) -> str:
    """Return word with symbols stripped from its ends."""
    return word.strip(string.punctuation)


def split_words(line: str) -> List[str]:
    """Return the list of words contained in a line."""
    # Normalize any camel cased words first.
    line = " ".join(w for w in _NORM_REGEX.split(line) if w)

    return [normalize(w) for w in _WORD_REGEX.split(line)]


def same_case(source: str, destination: str) -> str:
    """Return destination with same case as source."""
    if source and source[:1].isupper():
        return destination.capitalize()
    return destination


def get_a_line(filename: str, lineno: int) -> str:
    """Read a specific line from a file."""
    # Perhaps caching this would be nice, but assuming not an insane
    # number of misspellings.
    return open(filename, "r", encoding="utf-8").readlines()[lineno - 1].rstrip()


def esc_sed(raw_text: str) -> str:
    """Escape chars for a sed command on a shell command line."""
    return raw_text.replace('"', '\\"').replace("/", "\\/")


def esc_file(raw_text: str) -> str:
    """Escape chars for a file name on a shell command line."""
    return raw_text.replace("'", "'\"'\"'")



from .misspelling_detector import MisspellingDetector
from .misspelling_factory import MisspellingFactory
from .misspelling_file_detector import MisspellingFileDetector
from .misspelling_json_detector import MisspellingJSONDetector

__all__ = [
    "MisspellingDetector",
    "MisspellingFactory",
    "MisspellingFileDetector",
    "MisspellingJSONDetector",
]



import os
import pathlib
import sys
from codecs import StreamWriter
from typing import List, TextIO, Tuple, Union

from tap.tap import TapType

from .misspelling_interface import IMisspellingChecker
from .utils import esc_file, esc_sed, same_case, split_words


class MisspellingChecker(IMisspellingChecker):
    def check(self, filename: pathlib.Path) -> Tuple[List[Exception], List[List[Union[str, int, str]]]]:
        """
        Checks the files for misspellings.
        Returns:
          (errors, results)
          errors: List of system errors, usually file access errors.
          results: List of spelling errors - each tuple is filename,
                   line number and misspelled word.
        """
        errors = []
        results = []
        if not os.path.isdir(filename):
            try:
                with open(filename, "r", encoding="utf-8") as f:
                    line_ct = 1
                    for line in f:
                        for word in split_words(line):
                            if word in self._misspelling_dict or word.lower() in self._misspelling_dict:
                                results.append([filename, line_ct, word])
                        line_ct += 1
            except UnicodeDecodeError:
                pass
            except IOError as exception:
                errors.append(exception)
        return errors, results

    def get_suggestions(self, word: str) -> List[str]:
        """
        Returns a list of suggestions for a misspelled word.
        Args:
          word: The word to check.

        Returns:
          List of zero or more suggested replacements for word.

        """
        suggestions = set(self._misspelling_dict.get(word, [])).union(
            set(self._misspelling_dict.get(word.lower(), []))
        )
        return sorted(same_case(source=word, destination=w) for w in suggestions)

    def dump_corrections(self) -> List[List[str]]:
        """Returns a list of misspelled words and corrections."""
        results = []
        for bad_word in sorted(self._misspelling_dict.keys()):
            for correction in self._misspelling_dict[bad_word]:
                results.append([bad_word, correction])
        return results

    def print_result(self, filenames: List[pathlib.Path], output: StreamWriter) -> bool:
        """
        Print a list of misspelled words and their corrections.

        Return True if misspellings are found.

        """
        found = False

        for filename in filenames:
            errors, results = self.check(filename)
            for res in results:
                suggestions = ",".join(['"%s"' % w for w in self.get_suggestions(res[2])])
                output.write(f"{res[0]}:{res[1]}: {res[2]} -> {suggestions}\n")
                found = True

            for err in errors:
                sys.stderr.write("ERROR: %s\n" % err)
            output.flush()

        return found

    def export_result_to_file(self, filenames: List[pathlib.Path], output: TextIO) -> None:
        """
        Save the list of misspelled words and their corrections into a file.
        """

        for filename in filenames:
            _, results = self.check(filename)
            for res in results:
                output.write(
                    "{}:{}: {} -> {}\n".format(
                        res[0],
                        res[1],
                        res[2],
                        ",".join(['"%s"' % w for w in self.get_suggestions(res[2])]),
                    )
                )

    def output_sed_commands(self, parser: TapType, args: TapType, filenames: List[pathlib.Path]) -> None:
        """
        Output a series of portable sed commands to change the file.
        """
        if os.path.exists(args.script_output):
            # Emit an error is the file already exists in case the user
            # forgets to give the file - but does give source files.
            parser.error('The sed script file "%s" must not exist.' % args.script_output)

        with open(args.script_output, "w", encoding="utf-8") as sed_script:
            for filename in filenames:
                errors, results = self.check(filename)
                for res in results:
                    suggestions = self.get_suggestions(res[2])
                    if len(suggestions) == 1:
                        suggestion = suggestions[0]
                    else:
                        suggestion = self.suggestion_generator.get_suggestion(res[0], res[1], res[2], suggestions)
                    if suggestion != res[2]:
                        sed_script.write(
                            f'cp "{esc_file(res[0])}" "{esc_file(res[0])},"\n'
                            f'sed "{res[1]}s/{esc_sed(res[2])}/{esc_sed(suggestion)}/"'
                            f' "{esc_file(res[0])}" > "{esc_file(res[0])},"\n'
                            f'mv "{esc_file(res[0])}," "{esc_file(res[0])}"\n'
                        )
                for err in errors:
                    sys.stderr.write(f"ERROR: {err}\n" % err)



from pathlib import Path
from typing import List, Optional, Union

from tap import Tap


def to_path(files: Union[List[Path], List[str]]) -> List[str]:
    return files


class MisspellingArgumentParser(Tap):
    """
    Argument parser with the options for the misspellings command
    """

    def __init__(self):
        super().__init__(underscores_to_dashes=True)

    file_list: Optional[Path] = None  # File containing list of files to check
    misspelling_file: Optional[Path] = None  # File containing list of misspelled words and corrections
    json_file: Optional[Path] = None  # Json file containing misspelled words and corrections
    script_output: Optional[
        Path
    ] = None  # Create a shell script to interactively correct the files - script saved to the given file
    export_file: Optional[Path] = None  # Export the list of misspelled words into a file
    dump_misspelling: bool = False  # Dump the list of misspelled words
    version: bool = False  # Version of the misspellings package
    files: Optional[Union[List[Path], List[str]]] = None  # Files to check

    def configure(self) -> None:
        self.add_argument(
            "files",
            nargs="*",
            help="Files to check",
            type=to_path,
        )
